from ontobio.ontol_factory import OntologyFactory
from greent.service import Service

#TODO: LOOKUP all the terms that map to this... or use an ancestor call that doesn't require such stuff (i.e. that handles this)
GENETIC_DISEASE=('DOID:630','http://purl.obolibrary.org/obo/EFO_0000508')
#GENETIC_DISEASE='EFO:0000508'
MONOGENIC_DISEASE='DOID:0050177'

class Mondo(Service):
    
    """ A pragmatic class to query the mondo ontology. Until better sources emerge, we roll our own. """ 
    def __init__(self, context):
        ofactory = OntologyFactory()
        #This gets the remote owl which would be better in general, but is apparently not fully up to date
        # giving us a problem with Acute Alcohol Sensitivity which does not register as a genetic condition
        self.ont = ofactory.create('mondo')
        #self.ont = ofactory.create('./mondo.owl')
        #This seems to be required to make the ontology actually load:
        _ = self.ont.get_level(0)
        
    def get_doid(self,identifier):
        """We have an identifier, and we are going to use MONDO to try to convert it to a DOID"""
        upper_id = identifier.upper()
        obj_ids = self.get_mondo_id(upper_id)
        #Are any of the ids we get back a DOID?
        doids = []
        for obj_id in obj_ids:
            if obj_id.startswith('DOID:'):
                doids.append(obj_id)
        if len(doids) > 0:
            return doids
        #Didn't get anything, so get the xrefs and find DOIDS
        for obj_id in obj_ids:
            xref_ids = self.ont.xrefs(obj_id)
            for xref_id in xref_ids:
                if xref_id.startswith('DOID:'):
                    doids.append( xref_id )
        return doids

    def get_mondo_id(self,obj_id):
        """Given an id, find the main key(s) that mondo uses for the id"""
        if self.ont.has_node(obj_id):
            obj_ids = [obj_id]
        else:
            obj_ids = self.ont.xrefs(obj_id, bidirectional=True)
        return obj_ids

    def has_ancestor(self,obj, terms):
        """Given an object and a term in MONDO, determine whether the term is an ancestor of the object.
        
        The object is a KNode representing a disease.
        Some complexity arises because the identifier for the node may not be the id of the concept in mondo.
        the XRefs in mondo are checked for the object if it is not intially found, but this may return more 
        than one entity if multiple mondo entries map to the same.  

        Returns: boolean representing whether any mondo objects derived from the subject have the term as an
                         ancestor.
                 The list of Mondo identifiers for the object, which have the term as an ancestor"""
        #TODO: The return signature is funky, fix it.
        obj_id = obj.identifier
        obj_ids = self.get_mondo_id(obj_id)
        return_objects=[]
        for obj_id in obj_ids:
            ancestors = self.ont.ancestors(obj_id)
            for term in terms:
                if term in ancestors:
                    return_objects.append( obj_id )
        return len(return_objects) > 0, return_objects

    def is_genetic_disease(self,obj):
        """Checks mondo to find whether the subject has DOID:630 as an ancestor"""
        return self.has_ancestor(obj, GENETIC_DISEASE)

    def is_monogenic_disease(self,obj):
        """Checks mondo to find whether the subject has DOID:0050177 as an ancestor"""
        return self.has_ancestor(obj, MONOGENIC_DISEASE)

def test():
    m = Mondo()
    from reasoner.graph_components import KNode,KEdge
    alc_sens = KNode('OMIM:610251','D')
    print(m.is_genetic_disease(alc_sens))
    print('------')
    huntingtons = KNode('DOID:12858','D')
    print(m.is_genetic_disease(huntingtons))

if __name__ == '__main__':
    test()



import logging
import requests
import pprint
import urllib
from greent.service import Service
from greent.service import ServiceContext
from util import Text
from reasoner.graph_components import KNode,KEdge

class QuickGo(Service):

    def __init__(self, context):
        super(QuickGo, self).__init__("quickgo", context)

    def go_term_xontology_relationships(self, node):
        url = "{0}/QuickGO/services/ontology/go/terms/{1}/xontologyrelations".format (self.url, node.identifier)
        response = requests.get(url).json ()
        return [ ( self.get_edge (r, predicate=xrel['relation']),
                   KNode(xrel['id'], '?') ) for r in response['results'] for xrel in r['xRelations'] ]

def test ():
    q = QuickGo (ServiceContext.create_context ())
    r = q.go_term_xontology_relationships (KNode("GO:0032762", 'G'))
    pprint.pprint (r)

if __name__ == '__main__':
    test ()



import requests
import urllib
from greent.service import Service
from greent.mondo import Mondo
from reasoner.graph_components import KNode,KEdge
import logging

class Biolink(Service):
    def __init__(self, context): #url="https://api.monarchinitiative.org/api"):
        super(Biolink, self).__init__("biolink", context)
        #self.url = url
    def gene_get_disease(self, gene_node):
        """Given a gene specified as an HGNC curie, return associated diseases. """
        #TODO: we're assuming that gene_node.identifier is a valid curie for calling inti biolink - validate
        ehgnc = urllib.parse.quote_plus(gene_node.identifier)
        logging.getLogger('application').debug('          biolink: %s/bioentity/gene/%s/diseases' % (self.url, ehgnc))
        r = requests.get('%s/bioentity/gene/%s/diseases' % (self.url, ehgnc)).json()
        edge_nodes = [ ]
        #TODO:  Do I just want to suck in everything?  It's probably smarter, but for now it's mostly nulls
        #       and there's some stuff I'm completely unclear on (evidence graph).  In the long run, though,
        #       probably yes.
        for association in r['associations']:
            if 'publications' in association and association['publications'] is not None:
                pubs = [ {'id': pub['id']} for pub in association['publications'] ]
            else:
                pubs = []
            obj = KNode(association['object']['id'], 'D', association['object']['label'] )
            rel = { 'typeid': association['relation']['id'], 'label':association['relation']['label'] }
            props = { 'publications': pubs, 'relation':rel }
            edge = KEdge( 'biolink', 'queried', props )
            edge_nodes.append( (edge , obj ) )
        #TODO: WARN if no edges found?
        #TODO: DEBUG number of edges found / query.id
        #print (edge_nodes)
        return edge_nodes
    def gene_get_genetic_condition(self, gene):
        """Given a gene specified as an HGNC curie, return associated genetic conditions.
        
        A genetic condition is specified as a disease that descends from a ndoe for genetic disease in MONDO."""

        disease_relations = self.gene_get_disease(gene)
        checker = Mondo()
        relations = []
        for relation, obj in disease_relations:
            is_genetic_condition, new_object_ids = checker.is_genetic_disease(obj)
            if is_genetic_condition:
                obj.properties['mondo_identifiers'] = new_object_ids
                relations.append( (relation,obj) )
        #print (" biolink relations %s" % relations)
        return relations

def test():
    """What do we get back for HBB"""
    relations = gene_get_disease(('HGNC:4827',))
    checker = Mondo()
    for p, a in relations:
        igc, nid = checker.is_genetic_disease(a)
        print(a['id'], igc, nid)

def test_output():
    ehgnc = urllib.parse.quote_plus("HGNC:6136")
    r = requests.get('https://api.monarchinitiative.org/api/bioentity/gene/%s/diseases' % ehgnc).json()
    import json
    with open('testbiolink.json','w') as outf:
        json.dump(r, outf, indent=4)


if __name__ == '__main__':
    test_output()



import requests
import json
from greent.service import Service
from greent.service import ServiceContext
from greent.util import LoggingUtil
from greent.util import Text
from reasoner.graph_components import KNode,KEdge

logger = LoggingUtil.init_logging (__name__)

class TranslatorKnowledgeBeaconAggregator(Service):
    """ Use Translator Knowledge Beacons via the aggregator. """

    def __init__(self, context):
        super(TranslatorKnowledgeBeaconAggregator, self).__init__("tkba", context)

    def request_concept (self, concept):
        url = "{0}/concepts?keywords={1}".format (self.url, Text.un_curie (concept.identifier))
        return requests.get (url).json ()

    def name_to_doid (self, name):
        result = []
        response = self.request_concept (name)
        seen = {}
        for r in response:
            for a in r['aliases']:
                if a.startswith ("DOID:"):
                    if not a in seen:
                        logger.debug ("      -- appending a {}".format (a))
                        result.append ( ( self.get_edge (r, predicate='synonym'), KNode(a, name.node_type) ) )
                        seen[a] = a
        return result

    def name_to_drugbank (self, name):
        response = self.request_concept (name)
        result = []
        seen = {}
        for r in response:
            for a in r['aliases']:
                if a.startswith ("DRUGBANK:"):
                    if not a in seen:
                        result.append ( ( self.get_edge (r, predicate='synonym'), KNode(a, name.node_type) ) )
                        seen[a] = a
        return list(set(result))
        
    def name_to_mesh (self, name):
        result = []
        response = self.request_concept (name)
        seen = {}
        for r in response:
            for a in r['aliases']:
                if a.startswith ("MESH:"):
                    if not a in seen:
                        result.append ( ( self.get_edge (r, predicate='synonym'), KNode(a, name.node_type) ) )
                        seen[a] = a
        return list(set(result))
    
    def name_to_mesh_disease (self, name):
        response = self.request_concept (name)
        is_a_disease = False
        for r in response:
            for a in r['aliases']:
                if a.startswith ("DOID:"):
                    is_a_disease = True
                    break
        result = []
        if is_a_disease:
            seen = {}
            for r in response:
                for a in r['aliases']:
                    if a.startswith ("MESH:"):
                        if not a in seen:
                            a = a.replace ("MESH:", "MESH.DISEASE:")
                            result.append ( ( self.get_edge (r, predicate='synonym'), KNode(a, name.node_type) ) )
                            seen[a] = a
        return list(set(result))

if __name__ == "__main__":
    t = TranslatorKnowledgeBeaconAggregator (ServiceContext.create_context ())
    print (t.name_to_mesh_disease (KNode("NAME.DISEASE:asthma", 'D')))
    print (t.name_to_doid (KNode("NAME.DISEASE:asthma", 'D')))




from problog.extern import problog_export

@problog_export('+str', '+str', '-str')
def concat_str(arg1, arg2):
    return arg1 + arg2

@problog_export('+str', '+str', '-str')
def call_service (arg1, arg2):
    return arg1 + arg2



import logging
from greent.triplestore import TripleStore
from greent.util import LoggingUtil
import unittest
from pprint import pprint

logger = LoggingUtil.init_logging (__file__)

class MeSH(object):
    def __init__(self, uri="http://id.nlm.nih.gov/mesh/sparql"):
        self.triplestore = TripleStore (uri)
    def get_broader (self, term):
        return self.triplestore.query_template (
            inputs={ "term" : term, "prefixes" : self.get_prefixes () },
            outputs= [ "obj", "name" ],
            template_text="""
            $prefixes
            SELECT DISTINCT ?obj ?name ?itemName FROM  <http://id.nlm.nih.gov/mesh>
            WHERE {
               ?item  meshv:broaderDescriptor ?obj ;
                      rdfs:label              ?itemName.
               ?obj   rdfs:label              ?name .
              filter (regex(lcase(str(?itemName)), lcase(str("$term"))))
            } 
            ORDER BY ?p
            """)

    """
    SELECT DISTINCT ?obj ?name FROM  <http://id.nlm.nih.gov/mesh>
            WHERE {
               $term   meshv:broaderDescriptor ?obj .
               ?obj    rdfs:label              ?name .
            } 
            ORDER BY ?p
            """
    def get_prefixes (self):
        return """
        PREFIX rdf:      <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX rdfs:     <http://www.w3.org/2000/01/rdf-schema#>
        PREFIX xsd:      <http://www.w3.org/2001/XMLSchema#>
        PREFIX owl:      <http://www.w3.org/2002/07/owl#>
        PREFIX meshv:    <http://id.nlm.nih.gov/mesh/vocab#>
        PREFIX mesh:     <http://id.nlm.nih.gov/mesh/>
        PREFIX mesh2015: <http://id.nlm.nih.gov/mesh/2015/>
        PREFIX mesh2016: <http://id.nlm.nih.gov/mesh/2016/>
        PREFIX mesh2017: <http://id.nlm.nih.gov/mesh/2017/>"""

    
class TestMeSH(unittest.TestCase):

    m = MeSH ()
    def test_get_broader (self):
        pprint (self.m.get_broader ("Asthma"))  #"mesh:D001249"))

if __name__ == '__main__':
    unittest.main ()




import os
import traceback
from greent.util import LoggingUtil
from pprint import pprint
from SPARQLWrapper import SPARQLWrapper2, JSON
from string import Template

logger = LoggingUtil.init_logging (__file__)
#import logging
#logger = LoggingUtil.init_logging (__file__, logging.DEBUG)


class TripleStore(object):
    """ Connect to a SPARQL endpoint and provide services for loading and executing queries."""

    def __init__(self, hostname):
        self.service =  SPARQLWrapper2 (hostname)

    def get_template (self, query_name):
        """ Load a template given a template name """
        return Template (self.get_template_text (query_name))

    def get_template_text (self, query_name):
        """ Get the text of a template given its name """
        query = None
        fn = os.path.join(os.path.dirname(__file__), 'query',
            '{0}.sparql'.format (query_name))
        with open (fn, 'r') as stream:
            query = stream.read ()
        return query
    
    def execute_query (self, query):
        """ Execute a SPARQL query.

        :param query: A SPARQL query.
        :return: Returns a JSON formatted object.
        """
        #print (query)
        self.service.setQuery (query)
        self.service.setReturnFormat (JSON)
        return self.service.query().convert ()
    
    def query (self, query_text, outputs, flat=False):
        """ Execute a fully formed query and return results. """
        response = self.execute_query (query_text)
        result = None
        if flat:
            result = list(map(lambda b : [ b[val].value for val in outputs ], response.bindings ))
        else:
            result = list(map(lambda b : { val : b[val].value for val in outputs }, response.bindings ))
        logger.debug ("query result: %s", result)
        return result

    def query_template (self, template_text, outputs, inputs=[]):
        """ Given template text, inputs, and outputs, execute a query. """
        return self.query (Template (template_text).safe_substitute (**inputs), outputs)
    
    def query_template_file (self, template_file, outputs, inputs=[]):
        """ Given the name of a template file, inputs, and outputs, execute a query. """
        return self.query (self.get_template_text (template), inputs, outputs)




import json
import requests
from service import Service

class OXO(Service):

    """ Generic id translation service. Essentially a highly generic synonym finder. """
    def __init__(self, context): #url="https://www.ebi.ac.uk/spot/oxo/api/search?size=500"):
        super(OXO, self).__init__("oxo", context)
        #self.url = url

    def request (self, url, obj):
        return requests.post (self.url,
                              data=json.dumps (obj, indent=2),
                              headers={ "Content-Type" : "application/json" }).json ()
    def query (self, ids):        
        return self.request (
            url = self.url,
            obj = {
                "ids"           : ids,
                "mappingTarget" : [],
                "distance"      : "2"
            })
    
    def mesh_to_other (self, mesh_id):
        """ Find connections from a mesh id to other vocabulary domains. """
        result = []
        response = self.query (ids=[ mesh_id ])
        searchResults = response['_embedded']['searchResults']
        if len(searchResults) > 0 and searchResults[0]['queryId'] == mesh_id:
            others = searchResults[0]['mappingResponseList']
            result = list(map(lambda v : v['curie'], others))
        return result
    
'''

Reference: a conversation with OXO:

General:

Request URL:http://www.ebi.ac.uk/spot/oxo/api/search?size=500
Request Method:POST
Status Code:200 OK
Remote Address:193.62.193.80:80
Referrer Policy:no-referrer-when-downgrade

Response Headers view source

Cache-Control:no-cache, no-store, max-age=0, must-revalidate
Content-Type:application/json;charset=UTF-8
Date:Mon, 25 Sep 2017 14:24:54 GMT
Expires:0
Pragma:no-cache
Server:Apache-Coyote/1.1
Strict-Transport-Security:max-age=0
Transfer-Encoding:chunked
X-Application-Context:application
X-Content-Type-Options:nosniff
X-Frame-Options:DENY
X-XSS-Protection:1; mode=block


Request Headers view source

Accept:application/json, text/javascript, */*; q=0.01
Accept-Encoding:gzip, deflate
Accept-Language:en-US,en;q=0.8
Connection:keep-alive
Content-Length:58
Content-Type:application/json; charset=UTF-8
Cookie:_ga=GA1.3.1876641436.1506348116; _gid=GA1.3.425452358.1506348116; _gat=1
Host:www.ebi.ac.uk
Origin:http://www.ebi.ac.uk
Referer:http://www.ebi.ac.uk/spot/oxo/search
User-Agent:Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.101 Safari/537.36
X-Requested-With:XMLHttpRequest

Query String Parameters view source

view URL encoded
size:500


Request Payload view source

{ids: ["MESH:D001249"], mappingTarget: [], distance: "2"}







RESPONSE:


{
  "_embedded" : {
    "searchResults" : [ {
      "queryId" : "MESH:D001249",
      "querySource" : null,
      "curie" : "MeSH:D001249",
      "label" : "Asthma",
      "mappingResponseList" : [ {
        "curie" : "MedDRA:10003553",
        "label" : "Neutrophilic asthma",
        "sourcePrefixes" : [ "UMLS" ],
        "targetPrefix" : "MedDRA",
        "distance" : 2
      }, {
        "curie" : "HP:0002099",
        "label" : "Asthma",
        "sourcePrefixes" : [ "HP" ],
        "targetPrefix" : "HP",
        "distance" : 1
      }, {
        "curie" : "UMLS:C3714497",
        "label" : "Reactive airway disease",
        "sourcePrefixes" : [ "HP", "UMLS" ],
        "targetPrefix" : "UMLS",
        "distance" : 2
      }, {
        "curie" : "OMIM:608584",
        "label" : "ASTHMA-RELATED TRAITS, SUSCEPTIBILITY TO, 2",
        "sourcePrefixes" : [ "EFO" ],
        "targetPrefix" : "OMIM",
        "distance" : 2
      }, {
        "curie" : "OMIM:611960",
        "label" : "ASTHMA-RELATED TRAITS, SUSCEPTIBILITY TO, 7",
        "sourcePrefixes" : [ "EFO" ],
        "targetPrefix" : "OMIM",
        "distance" : 2
      }, {
        "curie" : "SNOMEDCT:278517007",
        "label" : "Asthmatic bronchitis",
        "sourcePrefixes" : [ "BAO", "DOID", "UMLS" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "OMIM:600807",
        "label" : "ASTHMA, SUSCEPTIBILITY TO",
        "sourcePrefixes" : [ "EFO", "BAO", "DOID" ],
        "targetPrefix" : "OMIM",
        "distance" : 2
      }, {
        "curie" : "SNOMEDCT:266365004",
        "label" : "",
        "sourcePrefixes" : [ "BAO", "DOID" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "SNOMEDCT:187687003",
        "label" : "Asthma",
        "sourcePrefixes" : [ "BAO", "DOID", "UMLS" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "OMIM:611064",
        "label" : "ASTHMA-RELATED TRAITS, SUSCEPTIBILITY TO, 5",
        "sourcePrefixes" : [ "EFO" ],
        "targetPrefix" : "OMIM",
        "distance" : 2
      }, {
        "curie" : "EFO:0000270",
        "label" : "asthma",
        "sourcePrefixes" : [ "EFO" ],
        "targetPrefix" : "EFO",
        "distance" : 1
      }, {
        "curie" : "KEGG:05310",
        "label" : "",
        "sourcePrefixes" : [ "BAO", "DOID" ],
        "targetPrefix" : "KEGG",
        "distance" : 2
      }, {
        "curie" : "SNOMEDCT:155574008",
        "label" : "Asthma",
        "sourcePrefixes" : [ "BAO", "DOID", "UMLS" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "ICD10CM:J45.909",
        "label" : "Unspecified asthma, uncomplicated",
        "sourcePrefixes" : [ "BAO", "DOID" ],
        "targetPrefix" : "ICD10CM",
        "distance" : 2
      }, {
        "curie" : "SNOMEDCT:266398009",
        "label" : "Asthma: [NOS] or [attack]",
        "sourcePrefixes" : [ "BAO", "DOID" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "ICD10CM:J45.90",
        "label" : "Late onset asthma",
        "sourcePrefixes" : [ "BAO", "DOID" ],
        "targetPrefix" : "ICD10CM",
        "distance" : 2
      }, {
        "curie" : "NCIt:C28397",
        "label" : "Asthma",
        "sourcePrefixes" : [ "EFO", "BAO", "DOID", "UMLS" ],
        "targetPrefix" : "NCIt",
        "distance" : 2
      }, {
        "curie" : "ICD9CM:493.9",
        "label" : "",
        "sourcePrefixes" : [ "BAO", "DOID" ],
        "targetPrefix" : "ICD9CM",
        "distance" : 2
      }, {
        "curie" : "DOID:2841",
        "label" : "asthma",
        "sourcePrefixes" : [ "BAO", "DOID" ],
        "targetPrefix" : "DOID",
        "distance" : 1
      }, {
        "curie" : "SNOMEDCT:195967001",
        "label" : "Hyperreactive airway disease",
        "sourcePrefixes" : [ "EFO", "BAO", "DOID", "HP", "UMLS" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "SNOMEDCT:195979001",
        "label" : "",
        "sourcePrefixes" : [ "BAO", "DOID" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "ICD9CM:493",
        "label" : "Asthma",
        "sourcePrefixes" : [ "EFO", "BAO", "DOID", "UMLS" ],
        "targetPrefix" : "ICD9CM",
        "distance" : 2
      }, {
        "curie" : "SNOMEDCT:21341004",
        "label" : "Asthmatic bronchitis",
        "sourcePrefixes" : [ "BAO", "DOID", "UMLS" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "SNOMEDCT:155579003",
        "label" : "Asthma: [NOS] or [attack]",
        "sourcePrefixes" : [ "BAO", "DOID" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "ICD10CM:J45",
        "label" : "Asthma",
        "sourcePrefixes" : [ "BAO", "DOID", "UMLS" ],
        "targetPrefix" : "ICD10CM",
        "distance" : 2
      }, {
        "curie" : "UMLS:C0004096",
        "label" : "Asthma",
        "sourcePrefixes" : [ "UMLS" ],
        "targetPrefix" : "UMLS",
        "distance" : 1
      }, {
        "curie" : "SNOMEDCT:991000119106",
        "label" : "Reactive airway disease",
        "sourcePrefixes" : [ "HP" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "OMIM:MTHU003537",
        "label" : "Asthma",
        "sourcePrefixes" : [ "UMLS" ],
        "targetPrefix" : "OMIM",
        "distance" : 2
      }, {
        "curie" : "SNOMEDCT:195983001",
        "label" : "(Asthma: [exercise induced] or [allergic NEC] or [NOS]) or (allergic bronchitis NEC)",
        "sourcePrefixes" : [ "BAO", "DOID" ],
        "targetPrefix" : "SNOMEDCT",
        "distance" : 2
      }, {
        "curie" : "OMIM:607277",
        "label" : "ASTHMA-RELATED TRAITS, SUSCEPTIBILITY TO, 1",
        "sourcePrefixes" : [ "EFO" ],
        "targetPrefix" : "OMIM",
        "distance" : 2
      } ],
      "_links" : {
        "self" : {
          "href" : "http://www.ebi.ac.uk/spot/oxo/api/terms/MeSH:D001249"
        },
        "mappings" : {
          "href" : "http://www.ebi.ac.uk/spot/oxo/api/mappings?fromId=MeSH:D001249"
        }
      }
    } ]
  },
  "_links" : {
    "self" : {
      "href" : "http://www.ebi.ac.uk/spot/oxo/api/search"
    }
  },
  "page" : {
    "size" : 500,
    "totalElements" : 1,
    "totalPages" : 1,
    "number" : 0
  }
}
'''



import json
import os
import sys
import time
import unittest
from collections import defaultdict
from pprint import pformat
from pprint import pprint
from pprint import PrettyPrinter
from greent.client import GraphQL
from collections import namedtuple
from flask_testing import LiveServerTestCase

class TestTranslator(LiveServerTestCase):

    translator = GraphQL ("http://localhost:5001/graphql")    
    Translation = namedtuple ('Translation', [ 'thing', 'domain_a', 'domain_b', 'description' ])
    Translation.__new__.__defaults__ = (None, None, None, None, '')
    translations = [
        Translation ("Imatinib",     "http://chem2bio2rdf.org/drugbank/resource/Generic_Name", "http://chem2bio2rdf.org/uniprot/resource/gene",      "Drug->Target"),      
        Translation ("CDC25A",       "http://chem2bio2rdf.org/uniprot/resource/gene",          "http://chem2bio2rdf.org/kegg/resource/kegg_pathway", "Target->Pathway"), 
        Translation ("CACNA1A",      "http://chem2bio2rdf.org/uniprot/resource/gene",          "http://pharos.nih.gov/identifier/disease/name",      "Target->Disease"),     
        Translation ("Asthma",       "http://identifiers.org/mesh/disease/name",               "http://identifiers.org/mesh/drug/name",              "Disease->Drug"),              
        Translation ("DOID:2841",    "http://identifiers.org/doid",                            "http://identifiers.org/mesh/disease/id",             "DOID->MeSH"),             
        Translation ("MESH:D001249", "http://identifiers.org/mesh",                            "http://identifiers.org/doi",                         "MeSH->*")
    ]

    def create_app(self):
        from greent import app
        app = app.create_app(graphiql=True)
        app.config['TESTING'] = True
        app.config['LIVESERVER_PORT'] = 5001
        return app
    
    def test_translations (self):
        pp = PrettyPrinter (indent=4)
        for index, translation in enumerate (self.translations):
            name = "-- Translate {0} -> thing: {1} in domain {2} to domain {3}.".format (
                translation.description, translation.thing, translation.domain_a, translation.domain_b)
            with self.subTest (name=name):
                print ("{0} {1}".format (self.id (), name))
                pp.pprint (self.translator.translate (
                    thing=translation.thing,
                    domain_a=translation.domain_a,
                    domain_b=translation.domain_b))
                
    def run_test (self, translation):
        return self.translator.translate (thing=translation.thing, domain_a=translation.domain_a, domain_b=translation.domain_b)
    def find_test (self, name):
        result = None
        for t in self.translations:
            if t.description == name:
                result = t
                break
        return result
    def run_by_name (self, name):
        print ("Executing test: {0}".format (name))
        translation = self.find_test (name)
        print ("   -- Translate thing: {0} in domain {1} to domain {2}. {3}".format (
            translation.thing, translation.domain_a, translation.domain_b, translation.description))
        print (self.run_test (translation))
    
if __name__ == '__main__':
    unittest.main ()
    sys.exit (0)



import asyncio
import concurrent.futures
import requests
from collections import namedtuple
from greent.util import LoggingUtil
import multiprocessing

logger = LoggingUtil.init_logging (__name__)

Operation = namedtuple ('Operation', [ 'operation', 'arguments' ])
default_chunk_size = multiprocessing.cpu_count()

class AsyncUtil:
    
    @staticmethod
    async def parallel_requests (urls, process_response, degree=default_chunk_size):
        with concurrent.futures.ThreadPoolExecutor(max_workers=degree) as executor:
            loop = asyncio.get_event_loop ()
            futures = [ loop.run_in_executor (executor, requests.get, url) for url in urls ]
            for response in await asyncio.gather(*futures):
                process_response (response)

    @staticmethod
    def execute_parallel_requests (urls, response_processor, chunk_size=default_chunk_size):
        chunks = [urls[i:i + chunk_size] for i in range(0, len(urls), chunk_size)]
        for chunk in chunks:
            logger.debug (" async requests: {0}".format (len(chunk)))
            loop = asyncio.get_event_loop()
            loop.run_until_complete(AsyncUtil.parallel_requests(chunk, process_response=response_processor))

    @staticmethod
    async def parallel_operation (operations, process_response, degree=default_chunk_size):
        with concurrent.futures.ThreadPoolExecutor(max_workers=degree) as executor:
            loop = asyncio.get_event_loop ()
            futures = [ loop.run_in_executor (executor, op.operation, op.arguments) for op in operations ]
            for response in await asyncio.gather (*futures):
                process_response (response)
    @staticmethod
    def execute_parallel_operations (operations, response_processor, chunk_size=default_chunk_size):
        chunks = [operations[i:i + chunk_size] for i in range(0, len(operations), chunk_size)]
        for chunk in chunks:
            logger.debug (" async-op: {0}".format (len(chunk)))
            loop = asyncio.get_event_loop()
            loop.run_until_complete(AsyncUtil.parallel_operation(chunk, process_response=response_processor))



import os
from reasoner.graph_components import KEdge, KNode
from greent.config import Config
from greent.util import LoggingUtil

class ServiceContext:
    def __init__(self, config=None):
        self.config = Config (config if config else os.path.join (os.path.dirname (__file__), "greent.conf"))
    @staticmethod
    def create_context (config=None):
        return ServiceContext (config)
    
class Service:
    """ Basic characteristics of srvices. """

    def __init__(self, name, context):
        """ Initialize the service given a name and an application context. """
        self.context = context
        self.name = name
        self.url = context.config.get_service (self.name)["url"]

        setattr (self.context, self.name, self)
        
    def _type(self):
        return self.__class__.__name__
    
    def get_edge (self, props={}, predicate=None, pmids=[]):
        """ Generate graph edges in a standard way, propagating information needed for
        scoring and semantic context above. """
        if not isinstance (props, dict):
            raise ValueError ("Properties must be a dict")

        # Add a predicate describing the connection between subject and object.
        # Pass up pmids for provenance and confidence scoring.
#        print (pmids)
#        print ("props: **********************>>>>>>>>>>>> {}".format (props))
        props['stdprop'] = {
            'predicate' : predicate,
            'pmids'     : pmids
        }
        return KEdge (self.name, predicate, props)



import json
import logging
import operator
import os
import sys
import traceback
import unittest
import yaml
from enum import Enum
from greent.util import LoggingUtil
from greent.util import Resource
from greent.util import Text
from greent.util import DataStructure
from greent.service import Service
from greent.service import ServiceContext
from pprint import pformat
from neo4jrestclient.client import GraphDatabase,Relationship,Node

logger = LoggingUtil.init_logging (__file__, level=logging.DEBUG)

class TypeGraph(Service):
    def __init__(self, service_context):
        super (TypeGraph, self).__init__("rosetta-graph", service_context)
        url = "{0}/db/data/".format (self.url)        
        self.db = GraphDatabase(url)
        self.types = self.db.labels.create("Type")
        self.bio_type_metadata = {
            'Disease'          : { 'color' : 'Red'        },
            'Substance'        : { 'color' : 'Yellow'     },
            'Gene'             : { 'color' : 'Blue'       },
            'Pathway'          : { 'color' : 'Green'      },
            'Anatomy'          : { 'color' : 'Pink'       },
            'Phenotype'        : { 'color' : 'Gray'       },
            'GeneticCondition' : { 'color' : 'LightGreen' }
        }
        self.bio_types = {}
        logger.debug ("-- Initializing bio types.")
        for t in self.bio_type_metadata.keys ():
            self.bio_types[t] = self.db.labels.create (t)
        self.concepts = {}
    def set_concepts (self, concepts):
        for concept, instances in concepts.items ():
            for instance in instances:
                self.concepts[instance] = concept
    def get_concept (self, item):
        return self.concepts.get (item)
    def get_relationships (self, a, b):
        q = "MATCH (a:Type { name:'%s' })-[r]-(b:Type { name:'%s' }) return r".format (a, b)
        return self.db.query(q, returns=(client.Node, str, client.Relationship), data_contents=True)
    def find_or_create (self, name, iri=None):
        n = self.types.get(name=name)
        if len(n) == 1:
            n = n[0]
        elif len(n) > 1:
            raise ValueError ("Unexpected non-unique node: {}".format (name))
        else:
            n =  self.db.nodes.create (name=name, iri=iri)
            self.types.add (n)
            concept = self.get_concept(name)
            if concept:
                self.bio_types[concept].add (n)
        return n
    def set_node_property (self, node_name, key, value):
        node = self.db.node (name=node_name)
        node.set (key, value)        
    def add_edge (self, a, b, rel_name, predicate, op):
        a_node = self.find_or_create (a)
        b_node = self.find_or_create (b)
        a_rels = a_node.relationships.outgoing(rel_name, b_node)
        exists = False
        for rel in a_rels:
            if rel.properties.get ('op',None) == op:
                exists = True
        if not exists:
            enabled = predicate != "UNKNOWN"
            a_node.relationships.create(rel_name, b_node, predicate=predicate, op=op, enabled=enabled)
    def get_shortest_paths (self, a, b):
        return self.db.query (
            "MATCH (a:Type { name: '%s' }),(b:Type { name : '%s' }), p = allShortestPaths((a)-[*]-(b)) RETURN p" % (a,b),
            data_contents=True)
    def get (self, url):
        return requests.get(url).json ()
    def get_transitions_x (self, query):
        self.narratives = {
            "diseasename_to_phenotype" : {
                "cypher" :
                """
                MATCH (n{name:"NAME"})-[name_to_doid:DISEASENAME]->(d{name:"DOID"})-[doid_to_pharos:DOID_TO_PHAROS]->(pharos{name:"PHAROS"})-[pharos_to_hgnc:DISEASE_GENE]->(h{name:"HGNC"})-[hgnc_to_uberon:ANATOMY]->(a{name:"UBERON"})
                MATCH (h)-[hgnc_to_go:CELLCOMPONENT]-(ph{name:"GO"})
                RETURN n, "NAME", name_to_doid, d, "DOID", doid_to_pharos, pharos, "PHAROS", pharos_to_hgnc, h, "HGNC>0", hgnc_to_uberon, a, "HGNC>1", hgnc_to_go, ph"""
            },
            "drugname_to_pathway" : {
                "cypher" :
                """
                MATCH (n)-[name_to_drugbank:DRUGNAME]->(s{name:"DRUGBANK"})-[drugbank_to_uniprot:TARGETS]->(g{name:"UNIPROT"})-[uniprot_to_kegg:GENE_PATHWAY]->(p{name:"KEGG"})
                MATCH (g)-[synonym:SYNONYM]->(h{name:"HGNC"})
                RETURN n, name_to_drugbank, s, drugbank_to_uniprot, g, uniprot_to_kegg, p, synonym, h"""
            },
            "test_1": {
                "cypher" : """MATCH (n{name:"NAME"})-[*0..2]->(d:Disease)-[*0..2]->(g:Gene)-[*1..1]->(a:Anatomy) RETURN DISTINCT n,d,g,a"""
            },
            "test_2": {
                "cypher" :
                """MATCH (n:Type{name:'NAME'}), (d:Disease), p = (n)-[r*..2]-(d) RETURN p"""
            },
            "test_3" : {
                "cypher" :
                """MATCH (a:Type { name: 'NAME' }),(b:Gene), p = allShortestPaths((a)-[*]->(b)) WHERE ALL(x IN nodes(p)[1..-1] WHERE (x:Disease)) RETURN p"""
            },
            "name-to-anatomy" : {
                "cypher" :
                """MATCH (a{name:"NAME"}),(b:Anatomy), p = allShortestPaths((a)-[*]->(b)) RETURN p"""
            }
        }
        program = []
        result = self.db.query (query, data_contents=True)
        for row in result.rows[0]:
            print (json.dumps (row, indent=2))
            node_type = None
            for col in row:
                #print ("col-> {} {}".format (col, type(col)))
                if isinstance (col, str):
                    logger.debug ("graph transition builder: noting result type: {0}".format (col))
                    node_type = col.split('>')[0] if '>' in col else col
                elif isinstance(col, dict):
                    if 'name' in col:
                        logger.debug ("graph transition builder: noting result type: {0}".format (col))
                        node_type = col['name']
                    if 'op' in col:
                        logger.debug ("graph transition builder: is dict.")
                        op = col['op']
                        logger.debug ("  --and has op")
                        is_new = True
                        for level in program:
                            if level['node_type'] == node_type:
                                logger.debug ("  -- adding op {0} to level".format (op))
                                level['ops'].append (op)
                                is_new = False
                        if is_new:
                            logger.debug ("  -- creating program component for node type {0}".format (node_type))
                            program.append ({
                                'node_type' : node_type,
                                'ops'       : [ op ],
                                'collector' : [] 
                            })
        print ("output program {}".format (program))
        return program
    def get_transitions (self, a, b):
        result = []
        paths = self.get_shortest_paths (a, b)
        for r in paths:
            for p in r:
                L = self.get (p['nodes'][0])
                R = self.get (p['nodes'][1])
                L = self.db.node[int(L['metadata']['id'])]
                R = self.db.node[int(R['metadata']['id'])]
                for rel in p['relationships']:
                    rel_obj = self.get (rel)
                    if rel_obj['type'] == 'transition':
                        start = self.get (rel_obj['start'])
                        end = self.get (rel_obj['end'])
                        result.append ( (start['data']['name'], rel_obj['data']['op'], end['data']['name'] ) )
        logger.debug (" **> T:{}".format (result))
        return result
    def find_or_create_concept (self, concept, instances):
        concept_node = self.bio_types[concept].get (name=concept)
        if len(concept_node) == 1:
            logger.debug ("-- Loaded existing concept: {0}".format (concept))
            concept_node = concept_node[0]
        elif len(concept_node) > 1:
            raise ValueError ("Unexpected non-unique concept node: {}".format (concept))
        else:
            size = 32
            bio_type = self.bio_type_metadata[concept]
            logger.debug ("-- Creating concept {0} with instances {1}".format (concept, instances))
            concept_node = self.bio_types[concept].create (name=concept, color=bio_type['color'], width=size, height=size)
            for e in instances:
                other = self.find_or_create (name=e)
                concept_node.relationships.outgoing ('instance', other)
        return concept_node



import json
from flask import Flask
from flask_graphql import GraphQLView
from greent.schema import Schema
from flask.views import View
from flask import request
from flask import Response
from flask import current_app
from pyld import jsonld
from greent.util import LoggingUtil
from pprint import pprint

logger = LoggingUtil.init_logging (__file__)

class PatientStubView(View):
    ''' Fake static patient data for dev test environment. '''
    def dispatch_request(self):
        return """
        [
         {
          "birth_date": "2006-08-02 00:00:00",
          "diag": {"ICD10:B08.4": {"2016-07-08 00:00:00": "OUTPATIENT"},
                   "ICD9:V19.2": {"2006-12-19 00:00:00": "OUTPATIENT"}},
          "geoCode": {"GEO:LAT": "35.22056", "GEO:LONG": "-80.69664"},
          "medList": {"MDCTN:10427": "2016-03-10 00:00:00",
                      "MDCTN:9502": "2017-01-20 00:00:00"},
          "patient_id": "32227752",
          "race": "white",
          "sex": "M"
         },
         {
          "birth_date": "2006-08-02 00:00:00",
          "diag": {"ICD10:B08.4": {"2016-07-08 00:00:00": "OUTPATIENT"},
                   "ICD9:V19.2": {"2006-12-19 00:00:00": "OUTPATIENT"}},
          "geoCode": {"GEO:LAT": "35.22056", "GEO:LONG": "-80.69664"},
          "medList": {"MDCTN:10427": "2016-03-10 00:00:00",
                      "MDCTN:9502": "2017-01-20 00:00:00"},
          "patient_id": "32227753",
          "race": "white",
          "sex": "M"
         }
        ]
        """

class JSONLDView (GraphQLView):
    ''' Extend GraphQLView with ability to tag documents with JSON-LD context.
    https://antoniogarrote.wordpress.com/2016/11/23/graphql-is-for-silos/ '''
    def get_jsonld_context (self, context_id):
        context = None
        with open (context_id, "r") as stream:
            context = json.loads (stream.read ())
        return context
    def dispatch_request (self):
        response = super(JSONLDView, self).dispatch_request ()
        if isinstance (response, Response):
            context = self.get_jsonld_context ('greent/greent_context.json')
            doc = json.loads(response.get_data())
            expanded = jsonld.expand (doc['data'], { "expandContext" : context } )
            if True: #current_app.debug:
                logger.debug ("context: {}".format (json.dumps (context, indent=2)))
                logger.debug ("doc: {}".format (json.dumps (doc, indent=2)))
                logger.debug ("expanded: {}".format (json.dumps (expanded, indent=2)))
                logger.deug (" context-> {}".format (type(context)))
            logger.debug (" expanded-> {}".format (type(expanded)))
            #expanded = expanded[0]
            doc['data']['@context'] = context['@context']
            response.set_data(json.dumps (doc)) #json.dumps (expanded))
            #logger.debug (json.loads (response.get_data ()))            
        return response

def create_app(path='/graphql', **kwargs):
    app = Flask(__name__)
    app.debug = True
    app.add_url_rule(path, view_func=GraphQLView.as_view('graphql', schema=Schema, **kwargs))
    app.add_url_rule('/sgraphql', view_func=JSONLDView.as_view('sgraphql', schema=Schema, **kwargs))
    app.add_url_rule('/patients/', view_func=PatientStubView.as_view('patients'))
    return app

def main ():
    app = create_app(graphiql=True)
    app.run(host="0.0.0.0", threaded=True)

if __name__ == '__main__':
    main ()



import argparse
import json
import logging
import networkx as nx
import networkx.algorithms as nxa
import operator
import os
import sys
import traceback
import unittest
import yaml
from enum import Enum
from greent.async import AsyncUtil
from greent.util import LoggingUtil
from greent.util import Resource
from greent.util import Text
from greent.util import DataStructure
from greent.neo4j import Neo4JREST
from greent.service import Service
from greent.service import ServiceContext
from greent.graph import TypeGraph
from networkx.exception import NetworkXNoPath
from networkx.exception import NetworkXError
from pprint import pformat
from reasoner.graph_components import KNode,KEdge,elements_to_json
import requests
import requests_cache
from networkx.readwrite import json_graph
from neo4jrestclient.client import GraphDatabase,Relationship,Node

#requests_cache.install_cache('rosetta_cache')
logger = LoggingUtil.init_logging (__file__, level=logging.DEBUG)

class Translation (object):
    """ A translation is a conceptual container for some thing, its type, and an 
    object type to convert it to. """    
    def __init__(self, obj, type_a=None, type_b=None, description="", then=None):
        self.obj = obj
        self.type_a = type_a
        self.type_b = type_b
    def __repr__(self):
        return "Translation(obj: {0} type_a: {1} type_b: {2})".format (
            self.obj, self.type_a, self.type_b)

class Rosetta:
    """ Rosetta's translates between semantic domains generically and automatically.
    Based on a configuration file, it builds a directed graph where types are nodes.
    Types can be IRIs or CURIEs. Edges are annotated with the names of operators used
    to transition between the connected types. The engine can then accept requests to 
    translate a term from one domain to another. It does this by collecting transitions
    from the graph and executing the list of transitions. """
    
    def __init__(self, greentConf="greent.conf",
                 config_file=os.path.join (os.path.dirname (__file__), "rosetta.yml"),
                 override={},
                 init_db=False):
        """ Load the config file and set up a DiGraph representing the types we know 
        about and how to transition between them. """
        from greent.core import GreenT
        self.debug = False

        # Construct the GreenT core containing services.
        logger.debug ("-- Initialize GreenT core.")
        self.core = GreenT (config=greentConf, override=override)
        self.type_graph = TypeGraph (self.core.service_context)

        '''
        # Create a digraph.
        self.g = nx.DiGraph ()
        '''
        # Load the type graph schematic.
        logger.debug ("-- Loading Rosetta config file: {0}".format (config_file))
        with open (config_file, 'r') as stream:
            self.config = yaml.load (stream)
            
        # Prime the vocabulary
        logger.debug ("-- Initializing vocabulary.")
        self.curie = self.config["@curie"]
        self.to_curie_map = {}
        logger.debug ("-- Initializing Rosetta vocabulary")
        self.vocab = self.config["@vocab"]
        for k in self.vocab:
            #self.g.add_node (self.vocab[k])
            self.to_curie_map[self.vocab[k]] = k

        # Store the concept dictionary
        logger.debug ("-- Initializing Rosetta concept dictionary")
        self.concepts = self.config["@concepts"]

        # Build a curie map. import cmungall's uber context.
        uber = Resource.get_resource_obj (os.path.join ("jsonld", "uber_context.jsonld")) #os.path.join (os.path.dirname (__file__), "jsonld", "uber_context.jsonld")
        context = uber['@context']
        self.terminate (context)
        for k in context:
            self.curie[k] = context[k]
            self.vocab[k] = context[k]

        """ Create curie map. Initialize based on the Identifiers.org web API."""
        logger.debug ("-- Initializing curie map incorporating Identifiers.org vocabulary.")
        identifiers_org = Resource.get_resource_obj ('identifiers.org.json')
        for module in identifiers_org:
            curie = module['prefix'].upper ()
            url = module['url']
            self.curie[curie] = url
            self.to_curie_map[url] = curie
            self.vocab[curie] = url

        type_concepts = self.config['@concept']
        self.type_graph.set_concepts (type_concepts)
        
        # Exit if we're not creating the data shema.
        if not init_db:
            return

        """ Create graph nodes in our type graph. """
        for k, v in self.vocab.items ():
            if isinstance (v, str):
                self.type_graph.find_or_create (k, v)
        for concept, terminologies in type_concepts.items ():
            self.type_graph.find_or_create_concept (concept, terminologies)
        
        # Build the transition graph.
        logger.debug ("-- Initializing Rosetta transition graph.")
        transitions = self.config["@transitions"]
        '''
        errors = 0
        for L in transitions:
            for R in transitions[L]:
                if not L in self.vocab:
                    errors += 1
                    self.log_debug("{0} not in vocab.".format (L))
                    continue
                if not R in self.vocab:
                    errors += 1
                    self.log_debug ("{0} not in vocab.".format (R))
                    continue
                self.add_edge (L, R, data=transitions[L][R])
                self.add_edge (self.vocab[L], self.vocab[R], data=transitions[L][R])
                self.add_edge (self.vocab[L], R, data=transitions[L][R])
                self.add_edge (L, self.vocab[R], data=transitions[L][R])
        '''
        for L in transitions:
            for R in transitions[L]:
                assert L in self.vocab and R in self.vocab
                transition_dict = transitions[L][R]
                transition_obj = DataStructure.to_named_tuple ('TransitionTuple', transitions[L][R])
                if 'link' in transition_dict and 'op' in transition_dict:
                    self.type_graph.add_edge (L, R,
                                              rel_name=transition_obj.link.upper (),
                                              predicate=transition_obj.link.upper (),
                                              op=transition_obj.op)

        # Connect the Translator Registry
        logger.debug ("-- Connecting to translator registry.")
        subscriptions = self.core.translator_registry.get_subscriptions ()
        for s in subscriptions:
            t_a_iri = s[0]
            t_b_iri = s[1]
            method = s[2]
            op = "translator_registry.{0}".format (method["op"])
            if not 'link' in method or method['link'] == None:
                link='unknown' #continue
            link = link.upper ()
            #self.add_edge (t_a_iri, t_b_iri, op)
            t_a = self.make_up_curie (self.unterminate (t_a_iri))
            t_b = self.make_up_curie (self.unterminate (t_b_iri))
            if not t_a:
                logger.debug ("Unable to find curie for {}".format (t_b))
            elif not t_b:
                logger.debug ("Unable to find curie for {}".format (t_b))
            else:
                self.type_graph.find_or_create (t_a, iri=t_a_iri)
                self.type_graph.find_or_create (t_b, iri=t_b_iri)
                if link and op:
                    self.type_graph.add_edge (t_a, t_b, rel_name=link, predicate=link, op=op)

    def terminate (self, d):
        for k, v in d.items ():
            if isinstance(v, str) and not v.endswith ("/"):
                d[k] = "{0}/".format (v)                
    def unterminate (self, text):
        return text[:-1] if text.endswith ('/') else text
                
    def add_edge (self, L, R, data):
        if self.debug:
            logger.debug ("  +edge: {0} {1} {2}".format (L, R, data))
        self.g.add_edge (L, R, data=data)
        #self.type_graph.add_edge (L, R, data=data['op'])
        
    def guess_type (self, thing, source=None):
        """ Look for a CURIE we know. If that doesn't work, try one of our locally made up vocab words. """
        if thing and not source and ':' in thing:
            curie = thing.upper ().split (':')[0]
            if curie in self.curie:
                source = self.curie[curie]
        if source and not source.startswith ("http://"):
            source = self.vocab[source] if source in self.vocab else None
        return source

    def map_concept_types (self, thing, object_type=None):
        """ Expand high level concepts into concrete types our data sources understand. """

        # Try the CURIE approach.
        the_type = self.guess_type (thing.identifier) if thing and thing.identifier else None

        # If that didn't work, get candiddate types based on the (abstract) node type.
        if thing and not the_type:
            the_type = self.concepts.get (thing.node_type, None)
            if the_type:
                # Attempt to map them down to IRIs
                the_type = [ self.vocab.get(t,t) for t in the_type ]

        # Systematize this:
        # If the concept type is disease but the curie is NAME, we don't have a DOID.
        
        if isinstance(the_type,str):
            # If we've ended up with just one string, make it a list for conformity of return type
            the_type = [ the_type ]

        result = the_type if the_type else self.concepts.get (object_type, [ object_type ])

        curie = Text.get_curie (thing.identifier) if thing else None
        if curie:
            result = [ self.make_up_curie (curie) ] #[ self.vocab[curie] ]
            #result = [ self.vocab[curie] ]

        return result

    def get_translations (self, thing, object_type):
        """ 
        A Thing is a node with an identifier and a concept. The identifier could really be a number, an IRI, a curie, a proper noun, etc.
        A concept is a broad notion that maps to many specific real world types.
        Our job here is to do the best we can to figure out what specific type this thing is. i.e., narrower than a concept.
        1. One way to do that is by looking at curies. If it has one we know, use the associated IRI.
        2. If that doesn't work, map the concept to the known specific types associated with it,.
        3. Do this for thing a and thing b.
        4. We want to guess specifically via the curie if possible, to avoid a combinatoric explosion resulting from cross product-ing spurious types.
        """
        x_type_a = self.map_concept_types (thing, thing.node_type)
        x_type_b = self.map_concept_types (thing=None, object_type=object_type)
        logger.debug ("Mapped types: {0} : {1}".format (x_type_a, x_type_b))
        translations = [ Translation(thing, ta_i, tb_i) for ta_i in x_type_a for tb_i in x_type_b ] if x_type_a and x_type_b else []
        for t in translations:
            logger.debug ("%s", t)
        return translations

    def get_transitions0 (self, source, target):
        """ Consulting the type graph, ask for paths between the source and destination types. 
        Turn those paths into trasitions - operators to call to effect the speicifed conversions from source->target."""
        transitions = []
        if not source or not target:
            return transitions
        try:
            # Get shortest paths between source and destination.
            paths = nxa.all_shortest_paths (self.g, source=source, target=target)
            #paths = nxa.all_simple_paths (self.g, source=source, target=target)
            count = 0
            for path in paths:                
                count += 1
                logger.debug ("  path: {0}".format (path))

                # Create pairs of types to indicates steps in the path.
                steps = list(zip(path, path[1:]))
                logger.debug ("  steps: {}".format (steps))
                for step in steps:

                    # Get edges associated with this pari
                    edges = self.g.edges (step, data=True)
                    for e in edges:
                        
                        if e[0] == e[1]:
                            # For now, let's not do transitions to self.
                            continue
                        
                        if step[1] == e[1]:
                            # Collect the transition method for these types into our output.
                            logger.debug ("    trans: {0}".format (e))
                            transition = e[2]['data']['op']
                            transitions.append (( e[0], transition ))
            if count == 0:
                logger.debug ("No paths found between {0} and {1}".format (source, target))
        except NetworkXNoPath:
            #traceback.print_exc ()
            #self.debug_graph (source, target)
            pass
        except KeyError:
            pass
        return transitions

    def to_curie (self, text):
        return self.to_curie_map.get (text, None)
    
    def make_up_curie (self, text):
        """ If we got one, great. Yay for standards. If not, get creative. This is legitimate and important because we can't have
        useful automated reasoning without granular semantics. But folks make more specific sub domain names which is probably
        essential for semantics and therefore automation. Until we arrive at a better approach, lets accept this approach and make up
        a curie if the service author thought it was important to have one."""
        curie = self.to_curie (text)
        if not curie:
            pieces = text.split ('/')
            last = pieces[-1:][0]
            curie = last.upper ()
        return curie
    
    def get_transitions (self, source, target):
        return self.type_graph.get_transitions (
            self.to_curie (source),
            self.to_curie (target))
    def process_translations (self, subject_node, object_type):
        """ Given a subject node and object type, do whatever is involved in walking the graph from the subject
        to the object type. Figure out types. Get transitions. Execute the transitions. Aggregate and return results."""
        result = [ ]

        # Consult the graph and get valid translations. This is essentially the output of the cross product of
        # potential specific types derived from abstract types like 'D' or 'S'
        translations = self.get_translations (subject_node, object_type)

        # Execute the translations. Dig down to a lower level and actually execute the transitions.
        for index, translation in enumerate (translations):            
            #logger.debug ("{0} translation:".format (index))
            data = self.translate (thing=translation.obj,
                                   source=translation.type_a,
                                   target=translation.type_b)
            result += data if isinstance(data,list) else []
        return result

    def debug_graph (self, source, target):
        try:
            logger.debug ("Calculating all paths from {0} to {1}".format (source, target))
            all_paths = nxa.all_simple_paths (self.g, source=source, target=target)
            for p in all_paths:
                logger.debug ("  > from[{0}]: {1}".format (source, p[1:]))
        except Exception as e:
            logger.debug (traceback.format_exc ().split ("\n")[-2:-1])

    def get_ops (self, names):
        return operator.attrgetter(names)(self.core) if isinstance(names,str) else [ operator.attrgetter(n)(self.core) for n in names ]
        
    def translate (self, thing, source, target):
        """ Given a set of translation requests (go from A->B), get transitions (actual operators to execute A->B).
        Prime the response stack with the input node and a null edge - to be discarded before returning.
        Get transitions based on guessed types.
        For each transition, execute it against each node in the list at the top of the stack.
        This builds a new list which we add to the top of the stack. """

        # Yeah, we really need something to start with.
        if not thing:
            return None

        # Prime the stack.
        stack = [ [ ( None, thing ) ] ]

        # Consult the type graph to get a list of (potentially) useful transitions.
        transitions = self.get_transitions (
            source=self.guess_type (thing, source),
            target=self.guess_type (None, target))
        
        # Do the transitions.
        max_cycles = -1
        for transition in transitions:
            try:
                # Lookup the transition in the GreenT core.
                data_op = self.get_ops (transition[1]) #operator.attrgetter(transition[1])(self.core)

                # Top is the result of the last query invocation.
                top = stack[-1:][0] # stack[-1] is a slice of the stack list. stack[-1][0] is an alement of the stack list.
                
                # A bucket to store new results.
                new_top = []

                # Iterate over each result in the prior set.
                for cycle, i in enumerate (top):
                    node = i[1]
                    #self.log_debug ("       top: {}".format (top))
                    self.log_debug ("           > {0}({1}) in:{2} => ".format (transition[1], node, transition[0]), cycle)

                    if max_cycles > -1 and cycle > max_cycles:
                        break
                    
                    # Invoke the transition operator.
                    result = []
                    if isinstance (data_op, list):
                        for inner, op in enumerate(data_op):
                            try:
                                if max_cycles > -1 and inner > max_cycles:
                                    break
                                result += (op (node))
                            except:
                                logger.debug ("           Error invoking {0}".format (transition[1]))
                    else:
                        try:
                            result += data_op (node)
                        except:
                            logger.error ("Error invoking {0}".format (transition[1]))

                    result = [ i for op in data_op for i in op(node) ] if isinstance(data_op, list) else data_op (node)
                    new_top += result if result is not None else []

                    if result:
                        self.log_debug ("             response>: {0}".format (Text.short(result,150)), cycle)

                # Push the new result set to top of stack.
                stack.append (self.unique_by_node_id (new_top))
                #stack.append (new_top)
            except:
                traceback.print_exc ()
            
        return [ pair for level in stack for pair in level if isinstance(pair,tuple) and isinstance(pair[0], KEdge) ]
    
    def unique_by_node_id (self,seq):
        """ Make the list edge, node tuples unique with respect to nodes. If we find a reason to want duplicate nodes, change this. """
        seen = set()
        seen_add = seen.add
        return [ x for x in seq if not (x[1].identifier in seen or seen_add(x[1].identifier)) ]
    
    def log_debug (self, text, cycle=0, if_empty=False):
        if cycle < 3:
            if (text and len(text) > 0) or if_empty:
                logger.debug ("{}".format (text))

    def translate_levels (self, node, levels):
        results = []
        last = [ (None, node) ]
        max_cycles = 5
        for index, level in enumerate (levels.split ("/")):
            new_last = []
            for subindex, result in enumerate(last):
                if max_cycles > -1 and subindex > 5:
                    break
                new_last += self.process_translations (result[1], level)
            last = new_last
            results += last
        return results

    def graph (self, next_nodes, query):
        program = self.type_graph.get_transitions_x (query)
        print ("program: {}".format (program))
        if not program or len(program) == 0:
            return []
        primed = [ { 'collector' : next_nodes } ] + program
        for index, level in enumerate (program):
            logger.debug ("--Executing level: {0}".format (level))
            operators = level['ops']
            collector = level['collector']
            #logger.debug ("000000>-=-------------> {}".format (primed[index]['collector']))
            for edge_node in primed[index]['collector']:
                for operator in operators:
                    op = self.get_ops (operator)
                    try:
                        logger.debug ("--Invoking op {0}({1})".format (operator, edge_node[1].identifier))
                        results = op (edge_node[1])
                        logger.debug ("  result> {0}".format (Text.short (results)))
                        for r in results:
                            if index < len(program) - 1:
                                if not r[1].identifier.startswith (program[index+1]['node_type']):
#                                    raise ValueError ("Operator {0} is wired to return type: {1} but returned node with id: {2}".format (
#                                        operator, program[index+1]['node_type'], r[1].identifier))
                                    logger.debug ("Operator {0} is wired to return type: {1} but returned node with id: {2}".format (
                                        operator, program[index+1]['node_type'], r[1].identifier))
                        collector += results
#                        logger.debug ("--------> collector ---------- {}".format (collector))
#                        logger.debug ("--------> collector ---------- {}".format (json.dumps (primed, indent=2)))
                    except Exception as e:
                        traceback.print_exc()
                        logger.error ("Error processing {0}".format (operator))
        return [ edge_node for level in program for result in level['collector'] for edge_node in result ]
                    
if __name__ == "__main__":
    #from neo4j.v1.api import GraphDatabase
    '''
    from neo4j import *
    for x in neo4j:
        print (x)
#    from neo4j import GraphDatabase
    driver = GraphDatabase.driver("bolt://localhost:7687") #, auth=basic_auth("neo4j", "neo4j"))
    session = driver.session()    
    session.run("CREATE (a:Person {name: {name}, title: {title}})",
                {"name": "Arthur", "title": "King"})
    result = session.run("MATCH (a:Person) WHERE a.name = {name} "
                         "RETURN a.name AS name, a.title AS title",
                         {"name": "Arthur"})
    for record in result:
        print("%s %s" % (record["title"], record["name"]))
    session.close()

    sys.exit (0)
    '''
    parser = argparse.ArgumentParser(description='Rosetta.')
    parser.add_argument('--initialize-type-graph', help='Build the graph of types and semantic transitions between them.', action="store_true", default=False)
    parser.add_argument('-d', '--disease', help='A disease to analyze.', default=None)
    parser.add_argument('-s', '--drug', help='A drug to analyze.', default=None)
    args = parser.parse_args()
    
    translator = Rosetta (override={ 'async' : True }, init_db=args.initialize_type_graph)
    disease = []
    drug = []
    if args.disease:
        query = \
                """MATCH (a{name:"NAME.DISEASE"}),(b:Gene), p = allShortestPaths((a)-[*]->(b)) 
                WHERE NONE (r IN relationships(p) WHERE type(r)='UNKNOWN') 
                RETURN p"""
        disease = translator.graph ( [ ( None, KNode('NAME.DISEASE:{0}'.format (args.disease), 'D') ) ], query=query)
    if args.drug:
        query = \
                """MATCH (a{name:"NAME.DRUG"}),(b:Pathway), p = allShortestPaths((a)-[*]->(b)) 
                WHERE NONE (r IN relationships(p) WHERE type(r)='UNKNOWN') 
                RETURN p"""
        drug = translator.graph ([ ( None, KNode('NAME.DRUG:{0}'.format (args.drug), 'S') ) ], query=query)

    union = disease + drug
    blackboard = [ t[1].identifier for t in union ]
    print (blackboard)



from string import Template
import json
import os
import logging
from greent.service import Service
from greent.triplestore import TripleStore
from greent.util import LoggingUtil
from greent.util import Text
from reasoner.graph_components import KEdge, KNode
from pprint import pprint

logger = LoggingUtil.init_logging (__file__)

class ChemBioKS(Service):
    """ Generic service endpoints for medical and bio-chemical data. This set
        comprises portions of chem2bio2rdf (CTD, KEGG, PubChem, DRUGBANK) """

    def __init__(self, context): #triplestore):
        super(ChemBioKS, self).__init__("chembio", context)
        self.triplestore = TripleStore (self.url)

    def query_chembio (self, query):
        """ Execute and return the result of a SPARQL query. """
        return self.triplestore.execute_query (query)

    def get_exposure_conditions (self, chemicals):
        """ Identify conditions (MeSH IDs) triggered by the specified stressor
            agent ids (also MeSH IDs).

        :param chemicals: List of IDs for substances of interest.
        :type chemicals: list of MeSH IDs, eg. D052638
        """
        id_list = ' '.join (list(map (lambda d : "( mesh:{0} )".format (d),
                            chemicals)))
        text = self.triplestore.get_template ("ctd_gene_expo_disease").\
            safe_substitute (chemicals=id_list)
        results = self.triplestore.execute_query (text)
        return list(map (lambda b : {
            "chemical" : b['chemical'].value,
            "gene"     : b['gene'].value,
            "pathway"  : b['kegg_pathway'].value,
            "pathName" : b['pathway_name'].value,
            "pathID"   : b['pathway_id'].value,
            "human"    : '(human)' in b['pathway_name'].value
        },
                         results.bindings))

    def get_drugs_by_condition (self, conditions):
        """ Get drugs associated with a set of conditions.

        :param conditions: Conditions to find associated drugs for.
        :type conditions: List of MeSH IDs for conditions, eg.: D001249
        """
        if not isinstance (conditions,list):
            conditions = [ conditions ]

        conditions = list(map(lambda v : v.replace ("MESH:", "mesh:"), conditions))
        prefix = "mesh:"
        if any(map(lambda v : v.startswith(prefix), conditions)):
            prefix = ""
        condition_list = ', '.join (list(map (lambda d : " {0}{1} ".format (prefix, d) , conditions)))
        result = self.triplestore.query_template (
            inputs = { "diseaseIds" : condition_list.lower () },
            outputs = [ 'drugID', 'drugGenericName', 'pubChemCID', 'diseasePMIDs' ],
            template_text="""
            prefix mesh:           <http://bio2rdf.org/mesh:> 
            prefix ctd:            <http://chem2bio2rdf.org/ctd/resource/>
            prefix db_resource:    <http://chem2bio2rdf.org/drugbank/resource/>
            select ?drugID ?drugGenericName ?diseasePMIDs ?ctdChemDis ?pubChemCID where {
               values ( ?diseaseId ) { ( $diseaseIds ) }
               ?ctdChemDis  ctd:cid                        ?pubChemCID;
                            ctd:diseaseid                  ?diseaseId;
                            ctd:pubmedids                  ?diseasePMIDs.
               ?dbInter     db_resource:Name               ?name ;
	                    db_resource:DBID               ?drugID .
               ?drugID      db_resource:CID                ?pubChemCID ;
  	                    db_resource:Generic_Name       ?drugGenericName .
            }""")
        return result

    def get_drugs_by_condition_graph (self, conditions):
        drugs = self.get_drugs_by_condition (conditions.identifier)
        results = []
        for r in drugs:
            edge = KEdge ('c2b2r', 'conditionToDrug',
                          { 'cid' : r['pubChemCID'], 'pmids' : r['diseasePMIDs'] })
            node = KNode (r['drugID'].split('/')[-1:][0],
                          "http://chem2bio2rdf.org/drugbank/resource/drugbank_drug",
                          r['drugGenericName'])
            results.append ( (edge, node) )
        #logger.debug ("chembio drugs by condition: {}".format (results))
        return results

    def get_genes_pathways_by_disease (self, diseases):
        """ Get genes and pathways associated with specified conditions.

        :param diseases: List of conditions designated by MeSH ID.
        :return: Returns a list of dicts containing gene and path information.
        """
        diseaseMeshIDList = ' '.join (list(map (lambda d : "( mesh:{0} )".format (d), diseases)))
        text = self.triplestore.get_template ("genes_pathways_by_disease").safe_substitute (diseaseMeshIDList=diseaseMeshIDList)
        results = self.triplestore.execute_query (text)
        return list(map (lambda b : {
            "uniprotGene" : b['uniprotGeneID'].value,
            "keggPath"    : b['keggPath'].value,
            "pathName"    : b['pathwayName'].value,
            "human"       : '(human)' in b['pathwayName'].value
        },
        results.bindings))

    def get_drug_gene_disease (self, disease_name, drug_name):
        """ Identify targets and diseases assocaited with a drug name.
        :param disease_name: MeSH name of a disease condition.
        :type str: String
        :param drug_name: Name of a drug.
        :type str: String
        """
        text = self.triplestore.get_template ("drug_gene_disease").safe_substitute (
            diseaseName=disease_name,
            drugName=drug_name)
        results = self.triplestore.execute_query (text)
        return list(map (lambda b : {
            "uniprotSymbol" : b['uniprotSym'].value,
            "diseaseId"     : b['diseaseID'].value
        }, results.bindings))

    def drug_name_to_gene_symbol (self, drug_name):
        result = self.triplestore.query_template (
            inputs = { "drugName" : drug_name },
            outputs = [ 'uniprotSym', 'pmids', 'drugID' ],
            template_text="""
            prefix ctd:            <http://chem2bio2rdf.org/ctd/resource/>
            prefix db_resource:    <http://chem2bio2rdf.org/drugbank/resource/>
            select ?drugGenericName ?pmids ?drugID ?uniprotSym where {
               values ( ?drugName ) { ( "$drugName" ) }
               ?ctdChemGene ctd:cid                        ?pubChemCID;
                            ctd:pubmedids                  ?pmids;
                            ctd:gene                       ?uniprotSym .
               ?drugID      db_resource:CID                ?pubChemCID ;
  	                    db_resource:Generic_Name       ?drugGenericName .
               filter regex(lcase(str(?drugGenericName)), lcase(?drugName))
            }""")
        return list(map(lambda r : {
            'uniprotSym'   : r['uniprotSym'],
            'pmids'        : r.get('pmids', None),
            'drugID'       : r['drugID']
        }, result))

    def gene_symbol_to_pathway (self, uniprot_symbol):
        return self.triplestore.query_template (
            inputs = { "uniprotSymbol" : uniprot_symbol },
            outputs = [ "keggPath" ],
            template_text="""
            prefix kegg:           <http://chem2bio2rdf.org/kegg/resource/>
            prefix pharmgkb:       <http://chem2bio2rdf.org/pharmgkb/resource/>
            prefix ctd:            <http://chem2bio2rdf.org/ctd/resource/>
            select ?ctdGene ?uniprotID ?pathwayName ?keggPath where {
               values ( ?ctdGene ) { ( <$uniprotSymbol> ) }
               ?keggPath    kegg:protein    ?uniprotID ; kegg:Pathway_name ?pathwayName .
               ?pharmGene   pharmgkb:Symbol ?ctdGene ; pharmgkb:UniProt_Id ?uniprotID.
               ?ctdChemGene ctd:gene        ?ctdGene.
            } LIMIT 500
            """)

    def uniprot_to_hgnc (self, uniprot_symbol):
        return self.triplestore.query_template (
            inputs = { "uniprotID" : Text.un_curie (uniprot_symbol.identifier) },
            outputs = [ "hgncID" ],
            template_text="""
            prefix uniprot:    <http://chem2bio2rdf.org/uniprot/resource/gene/>
            prefix owl:        <http://www.w3.org/2002/07/owl#>
            prefix hgnc:       <http://chem2bio2rdf.org/rdf/resource/hgnc/>
            select distinct ?hgncID where {
               values ( ?uniprotID ) { ( uniprot:${uniprotID} ) }
               ?uniprotID <http://www.w3.org/2002/07/owl#sameAs> ?hgncID.
               filter ( strstarts (str(?hgncID), "http://bio2rdf.org/gene:"))
            }
            """)

    def graph_uniprot_to_hgnc (self, uniprot_symbol):
        result = uniprot_to_hgnc (uniprot_symbol)
        return [ ( self.get_edge (r, predicate='synonym'), KNode('HGNC:{0}'.format (r['hgncID'])) ) for r in result ]

    def graph_get_genes_by_disease (self, disease): #reasoner
        disease = disease.identifier.split (':')[1].lower ()
        response = self.get_genes_pathways_by_disease ([ disease ])
        results = []
        for r in response:
            edge = KEdge ('c2b2r', 'diseaseToGene', { 'keggPath' : r['keggPath'] })
            node = KNode ("UNIPROT:{0}".format (r['uniprotGene'].split('/')[-1:][0]),  "G")
            results.append ( (edge, node) )
        return results

    def graph_get_pathways_by_gene (self, gene): #reasoner        
        response = self.triplestore.query_template (
            inputs = { "gene" : gene.identifier.split(':')[1].upper () },
            outputs = [ 'keggPath' ],
            template_text="""
            prefix kegg:      <http://chem2bio2rdf.org/kegg/resource/>
            prefix drugbank:  <http://chem2bio2rdf.org/drugbank/resource/>
            prefix uniprot:   <http://chem2bio2rdf.org/uniprot/resource/gene/>
            prefix ctd:       <http://chem2bio2rdf.org/ctd/resource/>
            prefix mesh:      <http://bio2rdf.org/mesh:>
            select ?drugGenericName ?uniprotGeneID ?pathwayName ?keggPath where {
               ?keggPath    kegg:protein                ?swissProtID ;
                            kegg:Pathway_name           ?pathwayName .
               ?keggInter   kegg:cid                    ?pubchemCID .
               ?dbInter     drugbank:GeneBank_ID        ?geneBankID ;
                            drugbank:SwissProt_ID       ?swissProtID ;
                            drugbank:gene               ?uniprotGeneID .
               ?drugID      drugbank:CID                ?pubchemCID ;
                            drugbank:Generic_Name       ?drugGenericName .
               ?ctd_disease ctd:diseaseid               ?diseaseID ;
                            ctd:cid                     ?pubchemCID .
               values ( ?uniprotGeneID ) {
                  ( uniprot:$gene )
               }
            } LIMIT 2000""")
        results = []
        for r in response:
            edge = KEdge ('c2b2r', 'geneToPathway', {})
            node = KNode ("KEGG:{0}".format (r['keggPath'].split('/')[-1:][0]), "P")
            results.append ( (edge, node) )
        return results

    def graph_drugname_to_gene_symbol (self, drug_name_node):
        drug_name = Text.un_curie (drug_name_node.identifier)
        response = self.drug_name_to_gene_symbol (drug_name)
        results = []
        for r in response:
            edge = self.get_edge (r, predicate="targets")
            node = KNode ("UNIPROT:{0}".format (Text.path_last (r['uniprotSym'])), "G")
            results.append ( (edge, node) )
        return results
    
    def graph_name_to_drugbank (self, drug_name_node):
        drug_name = Text.un_curie (drug_name_node.identifier)
        response = self.drug_name_to_gene_symbol (drug_name)
        results = []
        for r in response:
            edge = self.get_edge (r, predicate="drugname")
            node = KNode ("DRUGBANK:{0}".format (Text.path_last (r['drugID'])), "S")
            results.append ( (edge, node) )
        return results

    def graph_get_pathways_by_gene (self, gene): #reasoner        
        response = self.triplestore.query_template (
            inputs = { "gene" : gene.identifier.split(':')[1].upper () },
            outputs = [ 'keggPath' ],
            template_text="""
            prefix kegg:      <http://chem2bio2rdf.org/kegg/resource/>
            prefix drugbank:  <http://chem2bio2rdf.org/drugbank/resource/>
            prefix uniprot:   <http://chem2bio2rdf.org/uniprot/resource/gene/>
            prefix ctd:       <http://chem2bio2rdf.org/ctd/resource/>
            prefix mesh:      <http://bio2rdf.org/mesh:>
            select ?drugGenericName ?uniprotGeneID ?pathwayName ?keggPath where {
               ?keggPath    kegg:protein                ?swissProtID ;
                            kegg:Pathway_name           ?pathwayName .
               ?keggInter   kegg:cid                    ?pubchemCID .
               ?dbInter     drugbank:GeneBank_ID        ?geneBankID ;
                            drugbank:SwissProt_ID       ?swissProtID ;
                            drugbank:gene               ?uniprotGeneID .
               ?drugID      drugbank:CID                ?pubchemCID ;
                            drugbank:Generic_Name       ?drugGenericName .
               ?ctd_disease ctd:diseaseid               ?diseaseID ;
                            ctd:cid                     ?pubchemCID .
               values ( ?uniprotGeneID ) {
                  ( uniprot:$gene )
               }
            } LIMIT 2000""")
        results = []
        for r in response:
            edge = KEdge ('c2b2r', 'geneToPathway', {})
            node = KNode ("KEGG:{0}".format (r['keggPath'].split('/')[-1:][0]), "P")
            results.append ( (edge, node) )
        return results

    def graph_drugbank_to_uniprot (self, drugbank):
        response = self.triplestore.query_template (
            inputs = { "drugID" : "DB{0}".format (Text.un_curie (drugbank.identifier)) },
            outputs = [ "uniprotGeneID" ],
            template_text = """
            prefix drugbank:      <http://chem2bio2rdf.org/drugbank/resource/>
            prefix drugbank_drug: <http://chem2bio2rdf.org/drugbank/resource/drugbank_drug/>
            prefix ctd:           <http://chem2bio2rdf.org/ctd/resource/>
            select distinct ?uniprotGeneID where {
               values ( ?drugID ) { ( drugbank_drug:${drugID} ) }
               ?dbInter     drugbank:GeneBank_ID        ?geneBankID ;
                            drugbank:gene               ?uniprotGeneID .
               ?drugID      drugbank:CID                ?pubchemCID ;
                            drugbank:Generic_Name       ?drugGenericName .
               ?ctd_disease ctd:diseaseid               ?diseaseID ;
                            ctd:cid                     ?pubchemCID .
            }""")
        return [ ( self.get_edge (r, predicate='targets'),
                   KNode ("UNIPROT:{0}".format (r['uniprotGeneID'].split('/')[-1:][0]), "G") ) for r in response ]

    def graph_diseasename_to_uniprot (self, disease):
        results = []
        response = self.triplestore.query_template (
            inputs = { "diseaseName" : Text.un_curie (disease.identifier) },
            outputs = [ "pubChemCID" ],
            template_text = """
            prefix ctd: <http://chem2bio2rdf.org/ctd/resource/>
            select distinct ?pubChemCID where {
               values ( ?diseaseName ) { ( "$diseaseName" ) }
               ?ctdChemDis  ctd:cid         ?pubChemCID;
                            ctd:diseasename ?diseaseNameRec.
               filter regex(lcase(str(?diseaseNameRec)), lcase(?diseaseName))
            } LIMIT 1""")
        if len(response) > 0: # This is a disease.
            response = self.triplestore.query_template (
                inputs = { "diseaseName" : Text.un_curie(disease.identifier) },
                outputs = [ "disPmids", "chemPmids", "uniprotSym" ],
                template_text = """
                prefix ctd: <http://chem2bio2rdf.org/ctd/resource/>
                select ?disPmids ?ctdChemDis ?chemPmids ?uniprotSym ?diseaseId where {
                  values ( ?diseaseName ) { ( "$diseaseName" ) }
                  ?ctdChemGene ctd:cid         ?pubChemCID;
                               ctd:pubmedids   ?chemPmids;
                               ctd:gene        ?uniprotSym.
                  ?ctdChemDis  ctd:cid         ?pubChemCID;
                               ctd:diseaseid   ?diseaseId;
                               ctd:diseasename ?diseaseNameRec;
                               ctd:pubmedids   ?disPmids.
                  filter regex(lcase(str(?diseaseNameRec)), lcase(?diseaseName))
                } LIMIT 500""")
            for r in response:
                chemPmids = r['chemPmids']
                disPmids = r['disPmids']
                pmids = chemPmids + "|" + disPmids
                edge = self.get_edge (r, predicate='caused_by', pmids=pmids),
                node = KNode ("UNIPROT:{0}".format (r['uniprotSym'].split('/')[-1:][0]), "G")
                results.append ( (edge, node) )
        return results

    def graph_diseaseid_to_uniprot (self, drugbank):
        response = self.triplestore.query_template (
            inputs = { "diseaseID" : Text.un_curie (drugbank.identifier).lower () },
            outputs = [ "uniprotGeneID" ],
            template_text = """
            prefix drugbank:      <http://chem2bio2rdf.org/drugbank/resource/>
            prefix drugbank_drug: <http://chem2bio2rdf.org/drugbank/resource/drugbank_drug/>
            prefix ctd:           <http://chem2bio2rdf.org/ctd/resource/>
            prefix mesh:          <http://bio2rdf.org/mesh:> 
            select distinct ?uniprotGeneID where {
               values ( ?diseaseID ) { ( mesh:d001249 ) }
               ?dbInter     drugbank:gene               ?uniprotGeneID .
               ?drugID      drugbank:CID                ?pubchemCID.
               ?ctd_disease ctd:diseaseid               ?diseaseID ;
                            ctd:cid                     ?pubchemCID .
            }""")
        return [ ( self.get_edge (r, predicate='targets'),
                   KNode ("UNIPROT:{0}".format (r['uniprotGeneID'].split('/')[-1:][0]), "G") ) for r in response ]



import requests
import json
from greent.service import Service
from greent.service import ServiceContext

class Neo4JREST(Service):
    """ Speak to Neo4J via REST. """
    def __init__(self, name, context):
        super(Neo4JREST, self).__init__(name, context)
        self.cypher_uri = "{0}/db/data/cypher".format (self.url)
        self.query_endpoint = "{0}/db/data/transaction".format (self.url)
        self.schema_endpoint = "{0}/db/data/schema/".format (self.url)

    def request (self, url, obj):
        """ Make a request and return response. """
        return requests.post (url = url,
                             data = json.dumps (obj, indent=2),
                             headers={ "Content-Type" : "application/json" }).json ()
        
    def query (self, query, labels=None, node_properties=None, kinds=[ 'node' ]):
        """ Format a query. """
        response = self.request (
            url = self.query_endpoint,
            obj = {
                "statements": [
                    {
                        "statement": query,
                        "resultDataContents": [
                            "row",
                            "graph"
                        ],
                        "includeStats": True
                }
                ]
            })
        #print (json.dumps (response, indent=2))
        if node_properties or labels:
            response = self.filter_nodes (response, labels, node_properties, kinds)
        return response

    def execute_cypher (self, statement):
        response = requests.post (
            url = self.cypher_uri,
            data = json.dumps({
                "statements" : [{
                    "statement" : statement
                }]
            }),
            headers = { "Content-Type" : "application/json" })

    def filter_nodes (self, response, labels=None, properties=['identifier'], kinds=['node']):
        nodes = []
        relationships = []
        for r in response.get('results',[]):
            for d in r.get('data',[]):
                print ("data---------{}".format (json.dumps (d, indent=2)))
                print ("d2--------> {}".format (d.get('graph',{}).get('nodes',[])))
                d2 = d.get('graph',{}).get('nodes',[])
                print ("kinds: {}".format (kinds))
                if kinds == None or 'node' in kinds:
                    print ("--------> {}".format (d.get('graph',{}).get('nodes',[])))
                    for n in d2: #d.get('graph',{}).get('nodes'):
                        print("_______________GOT ONE")
                        print (json.dumps (n, indent=2))
                        if labels != None:
                            if any (map (lambda b : b in n['labels'], labels)):
                                print (labels)
                                if properties:
                                    obj = {}
                                    for prop in properties:
                                        obj[prop] = n['properties'][prop]
                                    nodes.append (obj)
                                else:
                                    print ("-----------")
                                    nodes.append (n['properties'])
                if 'relationships' in kinds:
                    for r in d['graph']['relationships']:
                        relationships.append (r)
        return nodes



import json
import unittest
from collections import defaultdict
from greent.mesh import MeSH
from greent.neo4j import Neo4JREST
from pprint import pprint
from greent.util import LoggingUtil

logger = LoggingUtil.init_logging (__file__)

class Chemotext(Neo4JREST):
    def __init__(self, context): #url="http://chemotext.mml.unc.edu:7474"):
        super (Chemotext, self).__init__("chemotext", context)
        self.mesh = MeSH ()
        
    def term_to_term (self, A, of_type=None, limit=100):
        response = self.query (
            query="MATCH (d:Term)-[r1]-(a:Art)-[r2]-(t:Term) WHERE d.name='%s' RETURN d, r1, a, r2, t LIMIT %s" % (A, limit))
        response = self.filter_nodes (response, labels=['Term'], properties=['name', 'type'])

        # Use MeSH data - slow but richer
        if of_type != None and response != None:
            new_response = []
            groups = defaultdict (lambda:None)
            for r in response:
                groups[r['name']] = r                
            for thing in groups:
                #logger.debug (" --mesh-req: {0}".format (thing))
                broader = self.mesh.get_broader (thing)
                obj = groups[thing]
                for k, v in of_type.items ():
                    for b in broader:
                        if k in b and b[k] == v:
                            #print ("b k v {0} {1} {2}".format (b, k, v))
                            obj['category'] = b['name']
                            obj['id'] = b['obj']
                            new_response.append (obj)
            response = new_response
        return response
    
    def disease_name_to_drug_name (self, disease, limit=100):
        result = []
        response = self.query (
            query="MATCH (d:Term {type:'Disease', name: '%s' })-[r1]-(a:Art)-[r2]-(t:Term {isDrug:true}) RETURN d, r1, a, r2, t LIMIT %s" %
            (disease, limit))
        for r in response['results'][0]['data']:
            result.append (r['row'][4]['name'])
        return list(set(list(result)))

    def graph_disease_name_to_drug_name (self, disease, limit=100):
        result = []
        drug_names = self.disease_name_to_drug_name (disease, limit)
        for r in drug_names:
            result.append ( ( self.get_edge (props=r), KNode("DRUGBANK.NAME:{0}".format (r['name']), 'D') ) )
        return result
    
'''
class TestChemotext(unittest.TestCase):

    chemotext = Chemotext (GreenT())
    
    def test_disease_to_drug (self):
        pprint ("Disease name to drug name:")
        pprint (self.chemotext.disease_name_to_drug_name ("Asthma"))

    def test_term_to_term (self):
        pprint (self.chemotext.term_to_term ('Asthma')[:10])

    def test_term_to_term_of_type (self):
        pprint ("term to term of type")
        pprint (self.chemotext.term_to_term ('Asthma', of_type={'name': 'Respiratory Hypersensitivity' }))
'''
if __name__ == '__main__':
    unittest.main ()




import json
import logging
import requests
import traceback
import re
import os
import sys
import yaml
from jinja2 import Template
from collections import defaultdict
from collections import namedtuple
from greent.service import Service
from greent.service import ServiceContext
from greent.util import LoggingUtil
from greent.util import Resource
from greent.util import DataStructure
from pprint import pprint
from reasoner.graph_components import KNode,KEdge
from jsonpath_rw import jsonpath, parse
from greent.util import Text

logger = LoggingUtil.init_logging (__name__, level=logging.DEBUG)

def defaultdict_to_regular(d):
    """ Recursively convert a defaultdict to a dict. """
    if isinstance(d, defaultdict):
        d = {k: defaultdict_to_regular(v) for k, v in d.items()}
    return d

class TranslatorRegistry(Service):
    """ Interact with Translator services. """
    def __init__(self, context):
        """ Read the Translator Registry root document. For each listed API, read its
            metadata. Then consider each path, parameter, and output in detail, regisgtering
            template strings used for invoking each service. Create a mapping to facilitate lookups
            of invocation templates based on type pairs. Later, need semantics about what the 
            meaning of the transitions is. Revenge of the semantic web and stuff."""
        super(TranslatorRegistry, self).__init__("transreg", context)
        self.verbose = True #False
        self.punctuation = re.compile('[ \./:]+')

        # Use cached model
        self.op_map = Resource.get_resource_obj ("transreg.yml", format='yaml')
        if isinstance (self.op_map, dict):
            logger.debug ("Loaded cached copy of translator registry config.")
            return
        '''
        registry_map = os.path.join (os.path.dirname (__file__), "transreg.yml")
        if os.path.exists (registry_map):
            with open (registry_map, "r") as stream:
                logger.debug ("Loading cached copy of translator registry config")
                self.op_map = yaml.load (stream.read ())
                return
        '''
        
        # Dynamically generate model
        self.op_map = defaultdict(lambda:defaultdict(lambda:defaultdict(None)))
        url = "{0}/API_LIST.yml".format (self.url)
        registry = yaml.load (requests.get (url).text)
        apis = {}
        for context in registry['APIs']:
            metadata = context['metadata']
            api_name = metadata.split (os.sep)[0].replace (" ","")
            logger.debug ("API: {}".format (api_name))
            api_url = "{0}/{1}".format (self.url, metadata)
            model = yaml.load (requests.get (api_url).text)
            servers = model.get('servers', [])
            server = None
            if isinstance(servers,list) and len(servers) > 0:
                server = servers[0]['url']
            paths = model.get('paths', {})
            for path in paths:
                obj = paths[path]
                logger.debug ("path: {}".format (path))
                get = obj['get'] if 'get' in obj else {}
                logger.debug ("get: {}".format (get))
                for parameters in get.get('parameters',{}):
                    logger.debug ("param: {}".format (parameters))
                    if 'x-valueType' in parameters:
                        values_in = parameters.get('x-requestTemplate',{})
                        logger.debug ("x-valueType: {}".format (values_in))
                        for v in values_in:
                            in_type = v['valueType']
                            x_template = v['template']
                            logger.debug ("in_type: {}".format (in_type))
                            success_response = get.get('responses',{}).get('200',{})

                            json_ld_url = success_response.get('x-JSONLDContext',None)
                            json_ld = {}
                            if json_ld_url:
                                json_ld = requests.get(json_ld_url).json ()

                            for response_value in success_response.get('x-responseValueType',{}):
                                out_type = response_value['valueType']
                                logger.debug ("out_type: {}".format (out_type))
                                logger.debug ("  --api> {0} in: {1} out: {2}".format (api_name, in_type, out_type))
                                self.op_map[api_name][in_type][out_type] = {
                                    "op"       : path,
                                    "get_url"  : "{0}{1}".format (server, x_template),
                                    "out_type" : response_value.get('valueType', None),
                                    "obj_path" : response_value.get('path', None),
                                    "jsonld"   : json_ld
                                }
        # Cache model
        vanilla_op_map = defaultdict_to_regular (self.op_map)
        with open (registry_map, "w") as stream:
            logger.debug ("Cache copy of registry map")
            yaml.dump (vanilla_op_map, stream, default_flow_style=False)

    def path_to_method_name (self, path):
        return path.replace ("{","").replace ("}","").replace ("?","/").replace ("=","_").replace ("/","_")
    
#    def to_named_tuple (self, type_name, d):
#        return namedtuple(type_name, d.keys())(**d)
    def get_service_metadata (self, api_name, in_type, out_type):
        metadata = self.op_map.get(api_name,{}).get (in_type,{}).get (out_type,{})
        return DataStructure.to_named_tuple ('ServiceMetadata', metadata) if len(metadata) > 0 else None #self.to_named_tuple ('ServiceMetadata', metadata) if len(metadata) > 0 else None
    
    def get (self, api_name, in_type, out_type, v, method_name):
        """ Invoke a GET requests on the specified API with the template and object. """
        result = []
        if not api_name in [ 'biolink' ]:
            return result
        try:
            service_metadata = self.get_service_metadata (api_name, in_type, out_type)
            logger.debug ("* Executing translator registry method: {0} in: {1} out: {2} template: {3} value: {4} ".format (
                api_name, in_type, out_type, service_metadata.get_url, v.identifier))

            """ Build and run the request. """
            url = Template (service_metadata.get_url).render (input=v.identifier)
            logger.debug ("                   translatR:get: {0}".format (url))
            response = requests.get (url).json ()
            
            """ Parse the response. """
            path_components = service_metadata.obj_path.split ('.')
            level = response
            for component in path_components:
                item = level.get (component, None)
                level = item
                print ("-------------------> item: {} {}".format (component, item))
            predicate = "{0}.{1}/{2}".format (api_name, method_name, path_components)
            result = [ ( self.get_edge(response, predicate=path_components), KNode(level, self.abstract(out_type)) ) ]
        except Exception as e:
            exc_type, exc_value, exc_tb = sys.exc_info()
            exception_text = traceback.format_exception (exc_type, exc_value, exc_tb)
            #logger.error (exception_text)
        return result
    
    def abstract(self, obj):
        """ Given a very specific type, translate it "up" to a robokop type. """
        x_type = '?'
        if ':' in obj:
            curie = obj.split (':')[0]
            type_m = {
                'HP'       : 'PH',
                'OMIM'     : 'D',
                'DOID'     : 'D',
                'UNIPROT:' : 'G',
                'NCBIGene' : 'G'
            }
            x_type = type_m[curie] if curie in type_m else '?'
            
    def human_name (self, iri):
        """ Replace a few things to simplify the text. """
        for prefix in [
                "http://identifiers.org/",
                "http://biothings.io/" ]:
            iri = iri.replace (prefix, "")
        return self.punctuation.sub ('', iri)

    def get_method_name (self, api, in_type, out_type):
        """ Build a normalized name for the method based on its input and output IRIs """
        x_api = self.human_name (api)
        x_in_type = self.human_name (in_type)
        x_out_type = self.human_name (out_type)
        return "{0}__{1}_to_{2}".format (x_api, x_in_type, x_out_type)

    def add_method (self, cls, api, in_type, out_type):
        """ Dynamically create a method for this API to do a get.
        For now, if there's no response processor configured, we just don't bother calling the API. """

        # Construct an API name.
        method_name = self.get_method_name (api, in_type, out_type)

        # Define the new method.
        def new_method(self, v):
            """ A dynamically created method to perform a translation. If we can't find a response processor, for now, don't bother. """
            return self.get (api, in_type, out_type, v, method_name)

        # Now add the method to this class. 
        new_method.__doc__ = "convert from {0} to {1}".format (in_type, out_type)
        new_method.__name__ = method_name
        setattr(cls, new_method.__name__, new_method)
        return method_name

    def get_subscriptions (self):
        """ Provide enough information to subscribe translator services as part of the Rosetta translation scheme.
        This involves passing back source type, destination type, and the method executing the transition.
        Rosetta only needs the name of the method since it will look up the actual method to dispatch to dynamically. """
        rosetta_config = Resource.get_resource_obj ("rosetta.yml", format='yaml')
        semantics = rosetta_config['@translator-semantics']
        subscriptions = []
        for api in self.op_map:
            for in_type, in_params in self.op_map[api].items ():
                for out_type, out_vals in self.op_map[api][in_type].items ():
                    predicate = semantics.get (api,{}).get (in_type,{}).get (out_type,None)
                    #if not predicate:
                    #    predicate = '*-missing-*'
                    subscriptions.append ((
                        in_type,
                        out_type,
                        {
                            "link" : predicate,
                            "op"   : self.add_method (TranslatorRegistry, api, in_type, out_type)
                        }
                    ))
        return subscriptions
    
if __name__ == "__main__":
    """ Load the registry """
    treg = TranslatorRegistry (ServiceContext.create_context ())

    """ Generate subscriptions """
    subscriptions = treg.get_subscriptions ()

    r = treg.myvariantinfo__uniprot_to_hgvs(KNode ('UNIPROT:AKT1','G'))
    pprint (r)

